#!/root/development/tasks/venv/bin/python3
import time, socket, threading, multiprocessing
import logging, logging.handlers
from datetime import datetime, timedelta
from tasks_core import get_postgres_connection, run_linux_performance_counters, monitor_task, log_rotate_task, \
    drop_backup_task
import tasks_elasticsearch
import tasks_mariadb
import tasks_mongodb
import tasks_mssql
import tasks_mysql
import tasks_oracle
import tasks_postgres
import tasks_redis
import k8s_tasks_postgres
import tasks_core
import os


AVAILABLE_MODULES = [tasks_elasticsearch, 
                     tasks_redis, 
                     tasks_mongodb,
                     tasks_mssql, 
                     tasks_mysql,
                     tasks_oracle,
                     tasks_postgres,
                     tasks_mariadb,
                     k8s_tasks_postgres,
                     tasks_core
                     ]
log_file = f'{os.path.join(os.getcwd(), "logs")}/123cluster.log'
# Set up a rotating file handler with a maximum size of 1MB and backup count of 10
handler = logging.handlers.RotatingFileHandler(filename=log_file, maxBytes=1024 * 1024, backupCount=10)

# Configure the formatter for the handler
formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s', datefmt='%d-%b-%y %H:%M')
handler = logging.handlers.TimedRotatingFileHandler(log_file, when='midnight', backupCount=1)
handler.setFormatter(formatter)

# Create a logger instance and add the handler to it
logger = logging.getLogger()
logger.addHandler(handler)

logger.setLevel(logging.INFO)


def do_agent() -> None:
    logging.info('Agent: %s %s', datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                 socket.gethostbyname(socket.gethostname()))
    try:
        with open('agent_params.txt', 'r') as f:
            lines = f.readlines()
    except FileNotFoundError as error:
        logging.error('Agent: %s', str(error))
    else:
        try:
            agent_key = lines[1].strip().split('=')[1]
        except IndexError as error:
            logging.error('Agent: There is no agent key in file')
            agent_key = ''
        finally:
            f.close()

        try:
            conn_agent = get_postgres_connection()
            with conn_agent:
                with conn_agent.cursor() as cur_agent:
                    cur_agent.execute("SELECT app_123monitor.update_agent_status(%s,%s)",
                                      [socket.gethostbyname(socket.gethostname()), agent_key])
                    conn_agent.commit()
        except Exception as error:
            logging.error('Agent: %s', str(error))
    finally:
        threading.Timer(15, do_agent).start()


def do_monitor() -> None:
    logging.info('Analyzer: %s', datetime.now().strftime("%Y-%m-%d %H:%M:%S"))
    try:
        conn_monitor = get_postgres_connection()
        with conn_monitor:
            with conn_monitor.cursor() as cur_monitor:
                cur_monitor.execute("SELECT app_123monitor.analyzer()")
                conn_monitor.commit()
    except Exception as error:
        logging.error('Monitor (Analyzer): %s', str(error))
    finally:
        threading.Timer(60, do_monitor).start()


def wait_to_midnight() -> None:
    # get current time
    now = datetime.now()

    # calculate number of seconds until midnight
    midnight = now.replace(hour=0, minute=0, second=0, microsecond=0) + timedelta(days=1)
    seconds_until_midnight = (midnight - now).seconds

    # sleep until midnight
    time.sleep(seconds_until_midnight)


def wait_to_next_hour() -> float:
    now = datetime.now()
    next_time = now.replace(minute=0, second=0, microsecond=0) + timedelta(hours=1)
    seconds_until_next = (next_time - now).total_seconds()
    return seconds_until_next


def wait_to_next_quarter_hour() -> float:
    now = datetime.now()

    # Determining the next multiple 15-minute interval
    next_minute = (now.minute // 15 + 1) * 15
    if next_minute == 60:
        next_minute = 0
        next_time = now.replace(hour=now.hour + 1, minute=0, second=0, microsecond=0)
    else:
        next_time = now.replace(minute=next_minute, second=0, microsecond=0)

    # Calculating how many seconds to wait
    seconds_until_next = (next_time - now).total_seconds()

    return seconds_until_next


def wait_to_next_five_minutes() -> float:
    now = datetime.now()
    next_minute = (now.minute // 5 + 1) * 5
    if next_minute == 60:
        next_time = now.replace(hour=now.hour + 1, minute=0, second=0, microsecond=0)
    else:
        next_time = now.replace(minute=next_minute, second=0, microsecond=0)

    seconds_until_next = (next_time - now).total_seconds()
    return seconds_until_next


def wait_to_next_minute() -> float:
    now = datetime.now()
    next_time = now.replace(second=0, microsecond=0) + timedelta(minutes=1)
    seconds_until_next = (next_time - now).total_seconds()
    return seconds_until_next


def housekeeper() -> None:
    logging.info('housekeeper: %s', datetime.now().strftime("%Y-%m-%d %H:%M:%S"))
    try:
        with get_postgres_connection() as conn_logs:
            with conn_logs.cursor() as cur_logs:
                delete_logs = """SELECT app_123monitor.housekeeper();"""
                cur_logs.execute(delete_logs)
                conn_logs.commit()
    except Exception as error:
        logging.error('housekeeper: %s', str(error))
    finally:
        wait_to_midnight()
        threading.Timer(0, housekeeper).start()


def do_week_stats() -> None:
    logging.info('do_week_stats: %s', datetime.now().strftime("%Y-%m-%d %H:%M:%S"))

    try:
        with get_postgres_connection() as conn_stats:
            with conn_stats.cursor() as cur_stats:
                cur_stats.execute("""SELECT app_123monitor.week_stats();""")
                conn_stats.commit()
    except Exception as error:
        logging.error('do_week_stats: %s', str(error))

    threading.Timer(wait_to_next_hour(), do_week_stats).start()


def do_day_stats() -> None:
    logging.info('do_day_stats: %s', datetime.now().strftime("%Y-%m-%d %H:%M:%S"))
    try:
        with get_postgres_connection() as conn_stats:
            with conn_stats.cursor() as cur_stats:
                day_stats = """SELECT app_123monitor.day_stats();"""
                cur_stats.execute(day_stats)
                conn_stats.commit()
    except Exception as error:
        logging.error('do_day_stats: %s', str(error))

    delay = wait_to_next_quarter_hour()
    threading.Timer(delay, do_day_stats).start()


def do_six_hours_stats() -> None:
    logging.info('do_six_hours_stats: %s', datetime.now().strftime("%Y-%m-%d %H:%M:%S"))

    try:
        with get_postgres_connection() as conn_stats:
            with conn_stats.cursor() as cur_stats:
                cur_stats.execute("""SELECT app_123monitor.six_hours_stats();""")
                conn_stats.commit()
    except Exception as error:
        logging.error('do_six_hours_stats: %s', str(error))

    threading.Timer(wait_to_next_five_minutes(), do_six_hours_stats).start()


def do_hour_stats() -> None:
    logging.info('do_hour_stats: %s', datetime.now().strftime("%Y-%m-%d %H:%M:%S"))

    try:
        with get_postgres_connection() as conn_stats:
            with conn_stats.cursor() as cur_stats:
                hour_stats = """SELECT app_123monitor.hour_stats();"""
                cur_stats.execute(hour_stats)
                conn_stats.commit()
    except Exception as error:
        logging.error('do_hour_stats: %s', str(error))

    delay = wait_to_next_minute()
    threading.Timer(delay, do_hour_stats).start()

def execute_task_function(func_name, *args):
    """Search for a function in all known modules and execute it if found."""
    global AVAILABLE_MODULES
    for module in AVAILABLE_MODULES:
        func = getattr(module, func_name, None)  # Get function safely
        if callable(func):
            func(*args)  # Execute function with positional arguments
            return  # Exit after first successful execution

    raise ValueError(f"Function '{func_name}' not found in any module.")

def do_task(row, allotted_time) -> None:
    task_id = row[0]
    logging.info(f"starting task {task_id} ({row[1]}):")
    try:
        conn = get_postgres_connection()
        cur = conn.cursor()
        # task = eval(func[1])                   #func[1] is function name
        # task(func[2],task_id)                  #func[2] is function parameters
        execute_task_function(row[1], row[2], task_id)
        if task_id:
            cur.execute('''UPDATE "clusterCore_task"
                           SET status='done',
                               next_due_date = CASE
                                   WHEN schedule = '15 seconds' THEN now() + interval '15 seconds'
                                   WHEN schedule = '1 minute' THEN date_trunc('minute', now() + interval '1 minute')
                                   WHEN schedule = '5 minutes' THEN date_trunc('minute', now() + interval '5 minutes')
                                   WHEN schedule = '10 minutes' THEN date_trunc('minute', now() + interval '10 minutes')
                                   WHEN schedule = '15 minutes' THEN date_trunc('minute', now() + interval '15 minutes')
                                   WHEN schedule = '1 hour' THEN date_trunc('minute', now() + interval '1 hour')
                                   WHEN schedule = '6 hours' THEN date_trunc('hour', now() + interval '6 hours')
                                   WHEN schedule = '1 day' THEN date_trunc('day', now() + interval '1 day')
                                   ELSE next_due_date
                               END
                           WHERE task_id=%s''', [task_id])
            conn.commit()
            cur.execute(
                '''UPDATE "clusterCore_task" SET status='ready' WHERE status='done' AND schedule NOT IN ('NOW', 'SCHED') AND task_id=%s''',
                [task_id])
            conn.commit()

        conn.commit()
        if not allotted_time:
            logging.info(f"completed task {task_id} ({row[1]}):")
    except Exception as error:
        logging.error(f'do_task: {task_id}: {str(error)}')
        cur.execute('''UPDATE "clusterCore_task" SET status='done' WHERE task_id=%s''', [task_id])
        conn.commit()
        cur.execute(
            '''UPDATE "clusterCore_task" SET status='ready' WHERE status='done' AND schedule NOT IN ('NOW', 'SCHED') AND task_id=%s''',
            [task_id])
        conn.commit()


def join_process(ps) -> None:
    try:
        logging.info(f'join_process: {ps["task_id"]}: {ps["allotted_time"]} seconds')
        if ps['ps'].is_alive():
            ps['ps'].join(timeout=ps['allotted_time'])
            if ps['ps'].is_alive():
                ps['ps'].terminate()
                logging.warning('Task %s has been running for more than the allotted time %s sec and was terminated',
                                ps['task_id'], ps['allotted_time'])
            else:
                logging.info(f"completed task {ps['task_id']} ({ps['task']}):")

            update_task_status(ps['ps'], ps['task_id'], ps['allotted_time'])
        else:
            logging.info(f"completed task {ps['task_id']} ({ps['task']}):")
    except Exception as error:
        logging.error(f'join_process: {ps["task_id"]}: {str(error)}')


def update_task_status(ps: multiprocessing.Process, task_id: int, allotted_time) -> None:
    # while ps.is_alive():
    #     time.sleep(0.2)

    try:
        conn = get_postgres_connection()
        cur = conn.cursor()
        cur.execute('''UPDATE "clusterCore_task" SET status='ready' WHERE task_id=%s''', [task_id])
        cur.execute('''UPDATE "clusterCore_tasklog"  SET duration = %s WHERE task_id = %s AND created_date =(SELECT max(created_date) FROM "clusterCore_tasklog" WHERE task_id=%s)''', [allotted_time+1, task_id, task_id])
        conn.commit()
        cur.close()
        conn.close()
    except Exception as error:
        logging.error(f'update_task_status: {task_id}: {str(error)}')


def main():
    while True:
        conn = None

        try:
            conn = get_postgres_connection()
            cur = conn.cursor()
            cur.execute('''SELECT ct.task_id, ct.task, ct.parameters, rm.runtime_red_threshold
                           FROM "clusterCore_task" AS ct
                           LEFT JOIN app_123monitor.rp_monitors AS rm ON ct.task_id = rm.task_id WHERE
                           (schedule='NOW' OR schedule='SCHED' OR
                           (schedule='15 seconds' AND TO_CHAR(current_timestamp,'SS') IN ('00','01','14','15','16','29','30','31','44','45','46','59')) OR
                           (schedule IN ('1 minute','60 seconds') AND TO_CHAR(current_timestamp,'SS') IN ('00','01','59')) OR
                           (schedule='5 minutes' AND TO_CHAR(current_timestamp,'MI') IN ('00','05','10','15','20','25','30','35','40','45','50','55')) OR
                           (schedule='10 minutes' AND TO_CHAR(current_timestamp,'MI') IN ('00','10','20','30','40','50')) OR
                           (schedule='15 minutes' AND TO_CHAR(current_timestamp,'MI') IN ('00','15','30','45')) OR
                           (schedule='1 hour' AND TO_CHAR(current_timestamp,'MI') IN ('00')) OR
                           (schedule='6 hours' AND TO_CHAR(current_timestamp,'HH:MI') IN ('00:00','06:00')) OR
                           (schedule='1 day' AND TO_CHAR(current_timestamp,'HH:MI') IN ('00:00'))
                           )
                           AND status='ready' AND (next_due_date <= now() or next_due_date IS NULL) AND target_id IS NOT NULL ORDER BY ct.task_id;''')
            tasks_rows = cur.fetchall()

            time.sleep(0.25)
            # check if list is empty
            if tasks_rows:
                processes = []
                time.sleep(0.9)
                for row in tasks_rows:
                    task_id = int(row[0])
                    cur.execute('''UPDATE "clusterCore_task" SET status='running' WHERE task_id=%s''', [task_id])
                    conn.commit()

                    # allotted_time in seconds
                    allotted_time = int(row[3]) / 1000 if row[3] is not None else 10 if str(
                        row[1]) == 'monitor_task' else None
                    ps = multiprocessing.Process(target=do_task, args=(row, allotted_time,))
                    processes.append({'task_id': task_id, 'task': row[1], 'ps': ps, 'pid': ps.pid, 'allotted_time': allotted_time})
                    ps.start()

                for ps in processes:
                    if ps['allotted_time']:
                        done_process = threading.Event()
                        threading.Timer(0, join_process, args=(ps,)).start()
        except KeyboardInterrupt:
            logging.info('task_service stopped by Keyboard Interrupt')
            exit(0)
        except Exception as error:
            logging.error(f'task_service: {str(error)}')
        finally:
            if conn:
                conn.close()


if __name__ == '__main__':
    done_agent = threading.Event()
    threading.Timer(15, do_agent).start()

    done_monitor = threading.Event()
    threading.Timer(60, do_monitor).start()

    done_week_stats = threading.Event()
    threading.Timer(wait_to_next_hour(), do_week_stats).start()

    done_day_stats = threading.Event()
    threading.Timer(wait_to_next_quarter_hour(), do_day_stats).start()

    done_six_hours_stats = threading.Event()
    threading.Timer(wait_to_next_five_minutes(), do_six_hours_stats).start()

    done_hour_stats = threading.Event()
    threading.Timer(wait_to_next_minute(), do_hour_stats).start()

    # done_delete = threading.Event()
    # threading.Timer(0, housekeeper).start()

    main()
